/*
  Light Painting Stick
  Copyright 2012 Gustav Evertsson
  www.guzzzt.com - me@guzzzt.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <PString.h>
#include <HL1606stripPWM.h>
#include <SdFat.h>
#include <HardwareSPI.h>
#include <stdint.h>

// Debug output only when USB is connected to a computer
#define DEBUG(d) if(SerialUSB.isConnected() && (SerialUSB.getDTR() || SerialUSB.getRTS())) { SerialUSB.print(millis(), DEC); SerialUSB.print(" - "); SerialUSB.println(d); }
#define DEBUG_VAL(d, v, t) if(SerialUSB.isConnected() && (SerialUSB.getDTR() || SerialUSB.getRTS())) { SerialUSB.print(millis(), DEC); SerialUSB.print(" - "); SerialUSB.print(d); SerialUSB.println(v, t); }

// Classes to handle the SD card
HardwareSPI spi1(1);
Sd2Card card;
SdVolume volume;
SdFile root;
SdFile file;

// Classes to handle the HL1606 LED stip
HardwareSPI ledSpi(2);
HardwareTimer ledTimer(2);
HL1606stripPWM strip = HL1606stripPWM(64, &ledSpi, &ledTimer, 20); 



// The timeout between button presses in the menu
#define BUTTON_DEBOUNCE_TIME 200
unsigned long menu_debounce_timer = 0;

// The time the menu is visible after last button press
#define MENU_TIME 4000
unsigned long menu_timer = 0;

// Pins for user input
#define BUTTON_PIN 8
#define SPEED_PIN 15
#define MENU_UP_PIN 17
#define MENU_DOWN_PIN 16

// File name, names is imgX.bmp where X is a number 1 to MAX_FILE_NR
int selectedImg = 1;  // Current file
char fileName[20];
PString str(fileName, sizeof(fileName));
#define MAX_FILE_NR 32

// BMP header information
int bmpWidth, bmpHeight;
uint8_t bmpDepth, bmpImageoffset;

// Error codes, number of LEDs flashing on error
#define ERROR_CARD_INIT     1
#define ERROR_VOLUME_INIT   2
#define ERROR_ROOT_INIT     3
#define ERROR_FILE_OPEN     5
#define ERROR_FILE_HEADER   7

///////////////////////////////////////////////////////
//                    MAIN SETUP                     //
///////////////////////////////////////////////////////
void setup() {
  SerialUSB.begin();
  //SerialUSB.println("Type any char to start");
  //while (!SerialUSB.available());
  //SerialUSB.println();
  strip.begin();  

  spi1.begin(SPI_18MHZ, MSBFIRST, 0);
  if(!card.init(&spi1)) {
    errorOutput(ERROR_CARD_INIT);
    DEBUG("card.init failed");
    while (1);
  }
  if(!volume.init(&card)) {
    errorOutput(ERROR_VOLUME_INIT);
    DEBUG("volume.init failed");
    while (1);
  }
  if(!root.openRoot(&volume)) {
    errorOutput(ERROR_ROOT_INIT);
    DEBUG("root.openRoot failed");
    while (1);
  }

 
  pinMode(BUTTON_PIN, INPUT);
  pinMode(MENU_UP_PIN, INPUT);
  pinMode(MENU_DOWN_PIN, INPUT);
  pinMode(SPEED_PIN, INPUT_ANALOG);
  
 
}

///////////////////////////////////////////////////////
//                     MAIN LOOP                     //
///////////////////////////////////////////////////////
void loop() {
  
  if(digitalRead(BUTTON_PIN)) {
    if(loadBmp())
      bmpdraw();
    file.close();
  }
  if(digitalRead(MENU_UP_PIN) || digitalRead(MENU_DOWN_PIN))
    menu();

  
}

///////////////////////////////////////////////////////
//         MENU TO SELECT FILE ON SD CARD            //
///////////////////////////////////////////////////////
void menu() {
  DEBUG_VAL("menu() - start - selectedImg: ", selectedImg, DEC);
 
  for (uint8_t i=0; i < MAX_FILE_NR; i++) {
    boolean exists = fileExists(i+1);
    if(i+1 == selectedImg && exists)
      strip.setLEDcolorPWM(63 - i, 0, 0, 255);
    else if(i+1 == selectedImg && !exists)
      strip.setLEDcolorPWM(63 - i, 128, 128, 0);
    else if(exists)
        strip.setLEDcolorPWM(63 - i, 0, 255, 0);
    else
      strip.setLEDcolorPWM(63 - i, 255, 0, 0);
  }
  
  menu_timer = millis() + (MENU_TIME * 2);
  while(millis() < menu_timer) {
    int menuBtn = 0;
    if(digitalRead(MENU_UP_PIN))
      menuBtn = MENU_UP_PIN;
    else if(digitalRead(MENU_DOWN_PIN))
      menuBtn = MENU_DOWN_PIN;

    if(menuBtn) {
      if(millis() < menu_debounce_timer)
        continue;
      menu_debounce_timer = millis() + BUTTON_DEBOUNCE_TIME;

      menu_timer = millis() + MENU_TIME;
      if(fileExists(selectedImg))
        strip.setLEDcolorPWM(63 - (selectedImg-1), 0, 255, 0);
      else
        strip.setLEDcolorPWM(63 - (selectedImg-1), 255, 0, 0);
     
      if(menuBtn == MENU_UP_PIN && selectedImg < MAX_FILE_NR) {
        selectedImg++;
      }
      else if(menuBtn == MENU_DOWN_PIN && selectedImg > 1) {
        selectedImg--;
      }
    
      if(fileExists(selectedImg))
        strip.setLEDcolorPWM(63 - (selectedImg-1), 0, 0, 255);
      else
        strip.setLEDcolorPWM(63 - (selectedImg-1), 128, 128, 0);
    }
  }
  
  for (uint8_t i=0; i< 64; i++) {
    strip.setLEDcolorPWM(i, 0, 0, 0);
  }
  DEBUG_VAL("menu() - end - selectedImg: ", selectedImg, DEC);
}

///////////////////////////////////////////////////////
//         CHECK IF FILE EXIST ON SD CARD            //
///////////////////////////////////////////////////////
boolean fileExists(int nr) {
  str.begin();
  str.print("img");
  str.print(nr);
  str.print(".bmp");

  DEBUG(str);
  
  if(file.open(&root, fileName, O_READ)) {
    file.close();
    return true;
  }
  return false;
}



///////////////////////////////////////////////////////
//                LOAD CURRENT BMP FILE              //
///////////////////////////////////////////////////////
boolean loadBmp() {
  str.begin();
  str.print("img");
  str.print(selectedImg);
  str.print(".bmp");

  DEBUG(str);
  
  if(!file.open(&root, fileName, O_READ)) {
    DEBUG("file.open failed");
    errorOutput(ERROR_FILE_OPEN);
    return false;
  }
  
  if (!bmpReadHeader()) { 
    DEBUG("bad bmp file");
    errorOutput(ERROR_FILE_HEADER);
    return false;
  }

  DEBUG_VAL("bmpWidth: ", bmpWidth, DEC);
  DEBUG_VAL("bmpHeight", bmpHeight, DEC);
  return true;
}

///////////////////////////////////////////////////////
//                  MAIN DRAW FUNCTION               //
///////////////////////////////////////////////////////
#define BUFFPIXEL 20  
void bmpdraw() {
  int speed = analogRead(SPEED_PIN);
  speed = map(speed, 0, 4096, 0, 75); 
  DEBUG_VAL("Speed ", speed, DEC);
  file.seekSet(bmpImageoffset);
  
  uint32_t time = millis();
  uint16_t p; 
  uint8_t g, b;
  int i, j;
  
  uint8_t sdbuffer[3 * BUFFPIXEL]; // 3 * pixels to buffer
  uint8_t buffidx = 3*BUFFPIXEL;
  
  
  for (i=0; i< bmpHeight; i++) {
    long lineStart = millis();
    for (j=0; j<bmpWidth; j++) {
      // read more pixels
      if (buffidx >= 3*BUFFPIXEL) {
        file.read(sdbuffer, 3*BUFFPIXEL);
        buffidx = 0;
      }
      
      // convert pixel from 888 to 565
      b = sdbuffer[buffidx++]; // blue
      g = sdbuffer[buffidx++]; // green
      r = sdbuffer[buffidx++]; // red
      /*
      p >>= 3;
      p <<= 6;
      
      g >>= 2;
      p |= g;
      p <<= 5;
      
      b >>= 3;
      p |= b;
      */
      //DEBUG_VAL("j ", j, DEC);
      //DEBUG_VAL("r 0x", r, HEX);
      //DEBUG_VAL("g 0x", g, HEX);
      //DEBUG_VAL("b 0x", b, HEX);
      
      
       strip.setLEDcolorPWM(j, r, g, b);

    }
    //DEBUG_VAL("Line ", i, DEC);
    while(lineStart + speed > millis());
  }
  
  for (uint8_t i=0; i< 64; i++) {
   strip.setLEDcolorPWM(i, 0, 0, 0);
  }
  DEBUG_VAL("Draw time in ms ", millis() - time, DEC);
}


///////////////////////////////////////////////////////
//     READ BMP FILE HEADER AND CHECK VALUES         //
///////////////////////////////////////////////////////
boolean bmpReadHeader() {
  // read header
  uint32_t tmp;
  
  if (read16() != 0x4D42) {
    DEBUG("magic bytes missing");
    return false;
  }
  
  // read file size
  tmp = read32(); 
  DEBUG_VAL("size 0x", tmp, HEX);
  
  // read and ignore creator bytes
  read32();
  
  bmpImageoffset = read32(); 
  DEBUG_VAL("offset ", bmpImageoffset, DEC);
  
  // read DIB header
  tmp = read32();
  DEBUG_VAL("DIB header size ", tmp, DEC);
  bmpWidth = read32();
  bmpHeight = read32();
  
  
  if (read16() != 1)
    return false;
  
  bmpDepth = read16();
  DEBUG_VAL("bitdepth ", bmpDepth, DEC);
  
  if (read32() != 0) {
    DEBUG("compression not supported");
    return false;
  }

  tmp = read32();
  DEBUG_VAL("the image data size ", tmp, DEC);
  
  return true;
}

///////////////////////////////////////////////////////
//                SHOW ERRROR MESSAGES               //
///////////////////////////////////////////////////////
void errorOutput(int error) {
  for (uint8_t x=0; x< 10; x++) {
    for (uint8_t i=0; i< 64; i++) {
      if(i >= 64 - error)
        strip.setLEDcolorPWM(i, 255, 0, 0);
      else
        strip.setLEDcolorPWM(i, 0, 0, 0);
    }
    delay(200);
    for (uint8_t i=0; i< 64; i++) {
      strip.setLEDcolorPWM(i, 0, 0, 0);
    }
    delay(200);
  }
}

///////////////////////////////////////////////////////
//             FILE READ - LITTLE ENDIAN             //
///////////////////////////////////////////////////////
uint16_t read16() {
  uint16_t d;
  uint8_t b;
  file.read(&b, 1);
  file.read(&d, 1);
  d <<= 8;
  d |= b;
  return d;
}

uint32_t read32() {
  uint32_t d;
  uint16_t b;
  
  b = read16();
  d = read16();
  d <<= 16;
  d |= b;
  return d;
}
